package cz.muni.csirt.analyza.simsearch;

/*-
 * #%L
 * Similarity Search Module
 * %%
 * Copyright (C) 2020 Masaryk University
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import cz.muni.csirt.analyza.client.RestClient;
import java.io.InputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.UUID;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.MediaType;

/**
 * Provides iterator of data file streams that satisfy the given filter.
 *
 * @author xbatko
 */
public class DataFileFilterIterator implements Iterator<UUID> {
    private final RestClient restClient;
    private final String restUri;
    private final Iterator<Map<String, ?>> dataFilesIterator;
    private Map<String, ?> lastDataFile;
    private final Client client;

    public DataFileFilterIterator(RestClient restClient, String restUri, String filter) {
        this.restClient = restClient;
        this.restUri = restUri;
        this.client = ClientBuilder.newClient();
        this.dataFilesIterator = client.target(restUri)
                .path("/generic-objects/filter")
                .queryParam("query", filter)
                .request(MediaType.APPLICATION_JSON)
                .get(new GenericType<Collection<Map<String, ?>>>() {}).iterator();
    }

    @Override
    public boolean hasNext() {
        return dataFilesIterator.hasNext();
    }

    @Override
    public UUID next() throws IllegalStateException {
        lastDataFile = dataFilesIterator.next();
        return getLastDataFileUUID();
    }

    public InputStream getLastDataFileStream() throws IllegalStateException {
        return restClient.getDataFileStream(getLastDataFileUUID());
    }

    private Object getProperty(Map<String, ?> genericObject, String propertyName) {
        if (propertyName == null)
            return null;
        Object properties = genericObject.get("properties");
        if (properties instanceof Collection)
            for (Map<String, ?> propertyMap : (Collection<Map<String, ?>>)properties) {
                if (propertyMap.get("propertyKey").equals(propertyName))
                    return propertyMap.get("valueString");
            }
        return null;
    }

    private Object getMapValue(Object map, String... names) {
        if (names != null)
            for (String name : names) {
                if (!(map instanceof Map))
                    return null;
                map = ((Map<?, ?>)map).get(name);
            }
        return map;
    }

    private UUID getUUIDValue(Object object) throws IllegalStateException {
        Object uuid = getMapValue(object, "uuid");
        if (uuid == null)
            throw new IllegalStateException("Object has no UUID: " + object);
        return UUID.fromString(uuid.toString());
    }

    public UUID getLinkedDataFile(String linkType, String propertyName, String propertyValue) {
        Collection<Map<String, ?>> links = client.target(restUri)
                .path("links/filter/" + getLastDataFileUUID())
                .request(MediaType.APPLICATION_JSON)
                .get(Collection.class);
        if (links != null)
            for (Map<String, ?> link : links) {
                if (linkType != null && !linkType.equals(getMapValue(link, "type", "type")))
                    continue;
                Object property = getProperty(link, propertyName);
                if (property != null && property.equals(propertyValue))
                    return getUUIDValue(getMapValue(link, "right"));
            }
        return null;
    }

    public UUID getLastDataFileUUID() throws IllegalStateException {
        return getUUIDValue(lastDataFile);
    }

    public Object getLastDataFileProperty(String key) {
        return lastDataFile == null ? null : lastDataFile.get(key);
    }

    public String getLastDataFilePropertyString(String key) {
        Object value = getLastDataFileProperty(key);
        if (value == null)
            return null;
        return value.toString();
    }
}
