package cz.muni.csirt.analyza.simsearch.extractor;

/*-
 * #%L
 * Similarity Search Module
 * %%
 * Copyright (C) 2020 Masaryk University
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import cz.muni.csirt.analyza.client.RestClient;
import cz.muni.csirt.analyza.entity.LinkType;
import cz.muni.csirt.analyza.simsearch.DataFileFilterIterator;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import messif.objects.extraction.ExtractorDataSource;
import messif.objects.extraction.Extractors;

/**
 * Implements the similarity search extractor module.
 * Running arguments are retrieved from the environment variables:
 * <ul>
 *   <li>{@link #FILTER_ENV_NAME} - the data filter for the warehouse</li>
 *   <li>{@link #QUERY_ENV_NAME} - the query object UUID in the warehouse</li>
 *   <li>{@link #DESCRIPTOR_ENV_NAME} - the name of the similarity descriptor to use</li>
 * </ul>
 * @author xbatko
 */
public class SimsearchExtractor {
    private final static String DW_URI_ENV = "DATAWAREHOUSE";
    private final static String MODULE_NAME = "simsearch_extractor";
    private final static String FILTER_ENV_NAME = MODULE_NAME.toUpperCase() + "_FILTER";
    private final static String QUERY_ENV_NAME = MODULE_NAME.toUpperCase() + "_QUERY";
    private final static String DESCRIPTOR_ENV_NAME = MODULE_NAME.toUpperCase() + "_DESCRIPTOR";
    private final static String DEFAULT_DESCRIPTOR = "vgg256";

    private final static String EXTRACTOR_EXEC_FORMAT = "extractor/extract_features.bin %s.model %1$s.proto -";

    public final static String LINK_TYPE_NAME = "derivedFrom";
    public final static String LINK_PROPERTY_DESCRIPTOR_NAME = "descriptor";

    /**
     * Main method of the extractor.
     * Reads the environment and runs the extraction.
     * @param args none
     * @throws Exception if there was an error running the extractor
     */
    public static void main(String[] args) throws Exception {
        String filter = System.getenv(FILTER_ENV_NAME);
        if (filter == null || filter.isEmpty())
            throw new IllegalArgumentException("No filter was specified");
        String query = System.getenv(QUERY_ENV_NAME);
        String descriptor = System.getenv(DESCRIPTOR_ENV_NAME);
        if (descriptor == null || descriptor.isEmpty())
            descriptor = DEFAULT_DESCRIPTOR;

        new SimsearchExtractor().process(filter, query, descriptor);
    }

    private final String restUri;
    private final RestClient restClient;

    public SimsearchExtractor() {
        this(System.getenv(DW_URI_ENV));
    }

    public SimsearchExtractor(String restUri) {
        this.restUri = restUri;
        this.restClient = new RestClient(restUri);
    }

    /**
     * Extracting method that reads all files that match the given filter and
     * (if specified) the given query with the given extractor.
     * @param filter the data filter for the warehouse
     * @param query the query object UUID in the warehouse
     * @param descriptor the name of the similarity descriptor to use
     * @throws IOException if there was an error reading/writing the data
     * @throws URISyntaxException if the rest API URI has wrong syntax
     */
    public void process(String filter, String query, String descriptor) throws IOException, URISyntaxException {
        String extractorPath = String.format(EXTRACTOR_EXEC_FORMAT, descriptor);
        process(
                new DataFileFilterIterator(restClient, restUri, filter),
                extractorPath, descriptor, Integer.MAX_VALUE
        );
        if (query != null && !query.isEmpty())
            process(
                    new DataFileFilterIterator(restClient, restUri, query),
                    extractorPath, descriptor, Integer.MAX_VALUE
            );
    }

    private InputStream transformData(InputStream stream, UUID uuid) throws IOException {
        try (BufferedReader reader = new BufferedReader(new InputStreamReader(stream))) {
            List<String> lines = new ArrayList<>();
            for (String line = reader.readLine(); line != null; line = reader.readLine())
                lines.add(line);
            int i = 3;
            String line;
            do {
                line = lines.get(i--);
            } while (line == null && i > 0);
            return new ByteArrayInputStream(("#objectKey messif.objects.keys.AbstractObjectKey " + uuid + "\n" + line + "\n").getBytes());
        }
    }

    /**
     * Extracting method that goes over the files in the given iterator and calls
     * the extractor if the descriptor is not found for that file.
     * @param iterator the iterator of the files to extract
     * @param extractorCmd the extractor command
     * @param descriptorName the name of the descriptor
     * @param maxCount the maximal number of files to extract
     * @throws IOException if there was an error reading/writing the data
     * @throws URISyntaxException if the rest API URI has wrong syntax
     */
    public void process(DataFileFilterIterator iterator, String extractorCmd, String descriptorName, int maxCount) throws IOException, URISyntaxException {
        LinkType linkType = new LinkType(LINK_TYPE_NAME);
        for (int count = 0; iterator.hasNext() && count < maxCount; count++) {
            UUID fileUUID = iterator.next();
            UUID descriptorUUID = iterator.getLinkedDataFile(LINK_TYPE_NAME, LINK_PROPERTY_DESCRIPTOR_NAME, descriptorName);
            if (descriptorUUID == null) {
                try (ExtractorDataSource dataSource = new ExtractorDataSource(iterator.getLastDataFileStream())) {
                    try (InputStream descriptorStream = Extractors.callExternalExtractor(extractorCmd, false, dataSource)) {
                        String caseId = iterator.getLastDataFilePropertyString("caseId");
                        UUID newDescriptorUUID = restClient.createDataFile(
                                transformData(descriptorStream, fileUUID),
                                fileUUID.toString() + "-" + descriptorName + ".data",
                                caseId
                        );
                        System.out.println("Extracted " + fileUUID + " into " + newDescriptorUUID);
                        UUID linkUUID = restClient.createLink(fileUUID, newDescriptorUUID, linkType, true, caseId);
                        restClient.updateUserProperties(linkUUID, LINK_PROPERTY_DESCRIPTOR_NAME, descriptorName);
                    }
                } catch (IOException e) {
                    System.err.println("Error extracting " + fileUUID + ": " + e);
                    //e.printStackTrace();
                }
            } else {
                System.out.println("Extracted " + fileUUID + " already present as " + descriptorUUID);
            }
        }
    }

    /**
     * Utility function that exports the descriptors of the objects with the given filter.
     * @param outputFile the name of the output file
     * @param filter the data warehouse filter
     * @param descriptorName the name of the descriptor
     * @throws IOException if there was an error reading/writing the data
     */
    public void export(String outputFile, String filter, String descriptorName) throws IOException {
        DataFileFilterIterator iterator = new DataFileFilterIterator(restClient, restUri, filter);
        try (FileOutputStream out = new FileOutputStream(outputFile)) {
            while (iterator.hasNext()) {
                UUID sourceFileUUID = iterator.next();
                UUID descriptorFileUUID = iterator.getLinkedDataFile(LINK_TYPE_NAME, LINK_PROPERTY_DESCRIPTOR_NAME, descriptorName);
                if (descriptorFileUUID != null) {
                    try (InputStream data = restClient.getDataFileStream(descriptorFileUUID)) {
                        // Copy data to output file
                        byte[] buffer = new byte[4096];
                        for (int bytesRead = data.read(buffer); bytesRead >= 0; bytesRead = data.read(buffer)) {
                            out.write(buffer, 0, bytesRead);
                        }
                    }
                } else {
                    System.out.println("No descriptor found for: " + sourceFileUUID);
                }
            }
        }
    }
}
