package cz.muni.csirt.analyza.simsearch.index;

/*-
 * #%L
 * Similarity Search Module
 * %%
 * Copyright (C) 2020 Masaryk University
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import cz.muni.csirt.analyza.simsearch.extractor.SimsearchExtractor;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import messif.objects.LocalAbstractObject;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.StreamGenericAbstractObjectIterator;

/**
 * Docker wrapper for starting the similarity search index.
 * Running arguments are retrieved from the environment variables:
 * <ul>
 *   <li>{@link #FILTER_ENV_NAME} - the data filter for the warehouse</li>
 *   <li>{@link #DESCRIPTOR_ENV_NAME} - the name of the similarity descriptor to use</li>
 * </ul>
 * @author xbatko
 */
public class SimsearchIndex {
    private final static String MODULE_NAME = "simsearch_index";
    private final static String FILTER_ENV_NAME = MODULE_NAME.toUpperCase() + "_FILTER";
    private final static String DESCRIPTOR_ENV_NAME = MODULE_NAME.toUpperCase() + "_DESCRIPTOR";
    private final static String DEFAULT_DESCRIPTOR = "vgg256";

    private final static int restApiPort = 8888;
    private final static File dataDir = new File("/usr/src/myapp/data/");
    private final static File dataFile = new File(dataDir, "index.data");
    private final static File pivotsFile = new File(dataDir, "pivots.data");
    private final static int defaultPivotCount = 16;
    private final static Class<? extends LocalAbstractObject> objectClass = messif.objects.impl.ObjectFloatVectorL2.class;

    private static String getEnvValue(String name) {
        String value = System.getenv(name);
        if (value == null)
            value = System.getenv(name.replace("_INDEX_", "_EXTRACTOR_"));
        return value;
    }

    public static void main(String[] args) throws IOException {
        String filter = getEnvValue(FILTER_ENV_NAME);
        if (filter == null || filter.isEmpty())
            throw new IllegalArgumentException("No filter was specified, expecting env " + FILTER_ENV_NAME);
        String descriptor = getEnvValue(DESCRIPTOR_ENV_NAME);
        if (descriptor == null || descriptor.isEmpty())
            descriptor = DEFAULT_DESCRIPTOR;
        
        dataDir.mkdir();
        new SimsearchExtractor().export(dataFile.getAbsolutePath(), filter, descriptor);
        int pivotCount = defaultPivotCount;
        try (StreamGenericAbstractObjectIterator<? extends LocalAbstractObject> iterator = new StreamGenericAbstractObjectIterator<>(objectClass, dataFile.getAbsolutePath())) {
            if (!iterator.hasNext())
                throw new IllegalArgumentException("Cannot create index: no data found for filter \"" + filter + "\"");
            AbstractObjectList<? extends LocalAbstractObject> pivots = iterator.getRandomObjects(pivotCount, false);
            try (OutputStream pivotFileStream = new FileOutputStream(pivotsFile)) {
                for (LocalAbstractObject pivot : pivots) {
                    pivot.write(pivotFileStream);
                }
            }
            pivotCount = pivots.size();
        }
        
        messif.utility.HttpApplication.main(new String[] {
            Integer.toString(restApiPort), "m-index.cf",
            "working-dir=" + dataDir.getAbsolutePath(),
            "dataclass=" + objectClass.getName(),
            "datafile=" + dataFile.getAbsolutePath(),
            "pivotcount=" + pivotCount,
            "pivotfile=" + pivotsFile.getAbsolutePath(),
            "buildMIndex"
        });
    }
}
