package cz.muni.csirt.analyza.simsearch.runsearch;

/*-
 * #%L
 * Similarity Search Module
 * %%
 * Copyright (C) 2020 Masaryk University
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import cz.muni.csirt.analyza.client.RestClient;
import cz.muni.csirt.analyza.entity.LinkType;
import java.io.IOException;
import java.util.Collection;
import java.util.Map;
import java.util.UUID;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.MediaType;

/**
 * Docker wrapper for starting the similarity search index.
 * Running arguments are retrieved from the environment variables:
 * <ul>
 *   <li>{@link #FILTER_ENV_NAME} - the data filter for the warehouse</li>
 *   <li>{@link #THRESHOLD_K_ENV_NAME} - the number of most similar objects to retrieve</li>
 *   <li>{@link #THRESHOLD_R_ENV_NAME} - radius of most similar objects to filter (negative value means unlimited)</li>
 * </ul>
 * @author xbatko
 */
public class SimsearchRunSearch {
    private final static String MODULE_NAME = "simsearch_knnsearch";
    private final static String QUERY_ENV_NAME = MODULE_NAME.toUpperCase() + "_QUERY";
    private final static String THRESHOLD_K_ENV_NAME = MODULE_NAME.toUpperCase() + "_THRESHOLD_K";
    private final static String THRESHOLD_R_ENV_NAME = MODULE_NAME.toUpperCase() + "_THRESHOLD_R";
    private final static String RESULT_LINK_ENV_NAME = MODULE_NAME.toUpperCase() + "RESULT_LINK";
    private final static String CASE_ID_ENV_NAME = MODULE_NAME.toUpperCase() + "CASE_ID";

    private final static String INDEX_MODULE_NAME = "simsearch_index";
    private final static String INDEX_ALGORITHM_ENDPOINT_ENV_NAME = "ENDPOINT_" + INDEX_MODULE_NAME.toUpperCase() + "_ALGORITHM";
    private final static String INDEX_HOSTNAME_ENV_NAME = "MODULE_ID_" + INDEX_MODULE_NAME.toUpperCase();
    private final static int DEFAULT_THRESHOLD_K = 10;
    private final static float DEFAULT_THRESHOLD_R = -1;
    private final static String DEFAULT_RESULT_LINK = "simsearch_result";
    private final static String DISTANCE_PROPERTY_NAME = "distance";

    public static void main(String[] args) throws IOException {
        String query = System.getenv(QUERY_ENV_NAME);
        if (query == null || query.isEmpty())
            throw new IllegalArgumentException("No query was specified, expecting env " + QUERY_ENV_NAME);
        UUID queryUuid = UUID.fromString(query);

        String thresholdK = System.getenv(THRESHOLD_K_ENV_NAME);
        int k = thresholdK == null || thresholdK.isEmpty() ? DEFAULT_THRESHOLD_K : Integer.parseInt(thresholdK);

        String thresholdR = System.getenv(THRESHOLD_R_ENV_NAME);
        float range = thresholdR == null || thresholdR.isEmpty() ? DEFAULT_THRESHOLD_R : Float.parseFloat(thresholdR);

        new SimsearchRunSearch().process(queryUuid, k, range);
    }

    private final String indexAlgorithmUri;
    private final Client client;
    private final RestClient restClient;
    private final LinkType resultLinkType;
    private final String caseId;

    public SimsearchRunSearch() {
        this(
            "http://" + System.getenv(INDEX_HOSTNAME_ENV_NAME) + ":" + System.getenv(INDEX_ALGORITHM_ENDPOINT_ENV_NAME) + "/",
            System.getenv(RESULT_LINK_ENV_NAME),
            System.getenv(CASE_ID_ENV_NAME)
        );
    }

    public SimsearchRunSearch(String indexAlgorithmUri, String resultLink, String caseId) {
        this.indexAlgorithmUri = indexAlgorithmUri;
        this.client = ClientBuilder.newClient();
        this.restClient = new RestClient(System.getenv("DATAWAREHOUSE"));
        this.resultLinkType = new LinkType(resultLink == null || resultLink.isEmpty() ? DEFAULT_RESULT_LINK : resultLink);
        this.caseId = caseId;
    }

    public void process(UUID queryUuid,  int k, float range) throws IOException {
        Collection<Map<String, ?>> results = client.target(indexAlgorithmUri)
                .path("search")
                .queryParam("locator", queryUuid.toString())
                .queryParam("k", k)
                .queryParam("range", range)
                .request(MediaType.APPLICATION_JSON)
                .get(new GenericType<Collection<Map<String, ?>>>() {});
        for (Map<String, ?> result : results) {
            linkQueryWithResult(queryUuid, toUUID(result.get("locator")), result.get("distance"));
        }
    }

    public static UUID toUUID(Object uuid) {
        if (uuid == null)
            return null;
        if (uuid instanceof UUID)
            return (UUID)uuid;
        return UUID.fromString(uuid.toString());
    }

    private UUID linkQueryWithResult(UUID queryUuid, UUID resultUuid, Object distance) {
        if (resultUuid == null)
            return null;
        UUID linkUUID = restClient.createLink(queryUuid, resultUuid, resultLinkType, true, caseId);
        if (distance != null)
            restClient.updateUserProperties(linkUUID, DISTANCE_PROPERTY_NAME, distance.toString());
        return linkUUID;
    }
}
